#include "TriMRFEnergy.h"
#include <time.h>





int TriMRFEnergy::optimize_CPU_RandOrder(Options &options)
{
	int itr = 0;
	bool is_done = false;
	std::vector<int> trav_order;
	randPermutation(_F, trav_order);

	cudaInitTimer() ;

	while ( itr++ < options.m_iterMax && !is_done) {
		is_done = true;
		std::cout << "iteration: " << itr << ".... "<< std::endl;		
		cudaBeginTimer ();
		for (int i = 0; i < _F ; i ++ ) {
			int t = trav_order[i];
			for (int v = 0; v < 3; v ++ ) {
				for (int l = 0; l < _L; l ++) {
					// Message M_{t->v; l}
					float min_v = f_array[t].min_jk(v, l);
					float dif = 0.5f * (v_array[f_array[t].vID[v]].theta_single[l] - min_v);

					v_array[f_array[t].vID[v]].theta_single[l] -= dif;
					f_array[t].add_msg(v, l, dif);
					if (dif > options.m_eps)
						is_done = false;
				}
			}
		}
		cudaEndTimer ();
	}

	cudaDestroyTimer ();
	return itr;
}

int TriMRFEnergy::optimize_CPU_RandPerFace(Options &options)
{
	int itr = 0;
	bool is_done = false;
	std::vector<int> trav_order;
	randPermutation(_F, trav_order);

	
	while ( itr++ < options.m_iterMax && !is_done) {
		is_done = true;
		std::cout << "iteration: " << itr << ".... "<< std::endl;		
		for (int i = 0; i < _F ; i ++ ) {
			int t = trav_order[i];
			// per face iteration
			{
				bool is_f_done = false;
				int itr_f = 0;
				while (!is_f_done) {
					is_f_done = true;
					//printf("iteration face (%d) : %d...\n", i, itr_f++);
					 {
						  for (int v = 0; v < 3; v ++ )
							  for (int l = 0; l < _L; l++) {
							// Message M_{t->v; l}
							float min_v = f_array[t].min_jk(v, l);
							float dif = 0.5f * (v_array[f_array[t].vID[v]].theta_single[l] - min_v);

							v_array[f_array[t].vID[v]].theta_single[l] -= dif;
							f_array[t].add_msg(v, l, dif);
							if (dif > options.m_eps) {
								is_done = false;
								is_f_done = false;
							}
						}
					}
				}
			}
		}
	}
	return 1;
}



int TriMRFEnergy::optimize_CPU_FixedOrder(Options &options)
{
	int itr = 0;
	bool is_done = false;

	while ( itr++ < options.m_iterMax && !is_done) {
		is_done = true;
		std::cout << "iteration: " << itr << ".... "<< std::endl;		
		for (int i = 0; i < _F ; i ++ ) {
			int t = i;
			for (int v = 0; v < 3; v ++ ) {
				for (int l = 0; l < _L; l++) {
					// Message M_{t->v; l}
					float min_v = f_array[t].min_jk(v, l);
					float dif = 0.5f * (v_array[f_array[t].vID[v]].theta_single[l] - min_v);

					v_array[f_array[t].vID[v]].theta_single[l] -= dif;
					f_array[t].add_msg(v, l, dif);
					if (dif > options.m_eps)
						is_done = false;
				}
			}
		}
	}
	return 1;
}


int TriMRFEnergy::optimize_GPU(Options &options)
{
#ifdef CUDA_OPTIMIZE
	maximalIndFaces();
	//cudaOptimizeTriMRF (_V, _F, _L, options.m_eps, options.m_iterMax);
	cudaOptimizeTriMRFNew (_V, _F, _L, options.m_eps, options.m_iterMax, ff_list);
	cudaGetEnergy(_V, _F, _L, arrayVert, arrayFace);
#endif
	return 1;
}


void TriMRFEnergy::maximalIndFaces()
{
	std::vector<int> VMarker, FMarker; // markers
	VMarker.reserve(_V);
	FMarker.reserve(_F);
	ff_list.clear();
	for (int i = 0; i < _V; i ++ ) VMarker.push_back(0);
	for (int i = 0; i < _F; i ++ ) FMarker.push_back(0);
	
	std::vector<int> f_list;
	
	int id1, id2, id3;
	int fcnt = 0; // count number of faces marked
	while (fcnt < _F) {
		f_list.clear();
		for (int i = 0; i < _V; i ++ ) VMarker[i] = 0;

		for (int f = 0; f < _F; f ++ ) {
			if (FMarker[f] == 1) 
				continue;
			
			id1 = fIdxArray[3*f + 0];
			id2 = fIdxArray[3*f + 1];
			id3 = fIdxArray[3*f + 2];
			if (VMarker[id1] == 0 && VMarker[id2] == 0 && VMarker[id3] == 0) {
				f_list.push_back(f);				
				VMarker[id1] = VMarker[id2] = VMarker[id3] = 1;
				FMarker[f] = 1;
				fcnt ++;
			}			
		}		
		ff_list.push_back(f_list);
	}

#ifdef TRIMRF_DEBUG
	for (int i = 0; i < ff_list.size(); i ++) {
		printf ("New List %d: \n", i);
		for (int j = 0; j < ff_list[i].size(); j ++) {
			printf("face %d : %d, %d, %d \n", j, fIdxArray[ff_list[i][j]*3+0], fIdxArray[ff_list[i][j]*3+1], fIdxArray[ff_list[i][j]*3+2]);
		}
	}
#endif
}